#include "btree_gist.h"


typedef struct macakey
{
	macaddr	  lower;
	macaddr	  upper;
}	MACAKEY;


/*
** macaddr ops
*/
PG_FUNCTION_INFO_V1(gmaca_compress);
PG_FUNCTION_INFO_V1(gmaca_union);
PG_FUNCTION_INFO_V1(gmaca_picksplit);
PG_FUNCTION_INFO_V1(gmaca_consistent);
PG_FUNCTION_INFO_V1(gmaca_penalty);
PG_FUNCTION_INFO_V1(gmaca_same);

Datum		gmaca_compress(PG_FUNCTION_ARGS);
Datum		gmaca_union(PG_FUNCTION_ARGS);
Datum		gmaca_picksplit(PG_FUNCTION_ARGS);
Datum		gmaca_consistent(PG_FUNCTION_ARGS);
Datum		gmaca_penalty(PG_FUNCTION_ARGS);
Datum		gmaca_same(PG_FUNCTION_ARGS);

static void gmaca_binary_union(Datum *r1, char *r2);
static int	mckey_cmp(const void *a, const void *b);


/* define for comparison */
#define MCGE( m1, m2 ) (DatumGetBool(DirectFunctionCall2( \
		macaddr_ge, \
		PointerGetDatum( m1 ), \
		PointerGetDatum( m2 ) \
)))
#define MCGT( m1, m2 ) (DatumGetBool(DirectFunctionCall2( \
		macaddr_gt, \
		PointerGetDatum( m1 ), \
		PointerGetDatum( m2 ) \
)))
#define MCEQ( m1, m2 ) (DatumGetBool(DirectFunctionCall2( \
		macaddr_eq, \
		PointerGetDatum( m1 ), \
		PointerGetDatum( m2 ) \
)))
#define MCLT( m1, m2 ) (DatumGetBool(DirectFunctionCall2( \
		macaddr_lt, \
		PointerGetDatum( m1 ), \
		PointerGetDatum( m2 ) \
)))
#define MCLE( m1, m2 ) (DatumGetBool(DirectFunctionCall2( \
		macaddr_le, \
		PointerGetDatum( m1 ), \
		PointerGetDatum( m2 ) \
)))



/**************************************************
 * macaddr ops
 **************************************************/

Datum
gmaca_compress(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
	GISTENTRY  *retval;

	if (entry->leafkey)
	{
		MACAKEY	   *r = (MACAKEY *) palloc(sizeof(MACAKEY));

		retval = palloc(sizeof(GISTENTRY));
                memcpy( &r->lower, (void*) DatumGetPointer(entry->key),sizeof(macaddr) );
                memcpy( &r->upper, (void*) DatumGetPointer(entry->key),sizeof(macaddr) );
		gistentryinit(*retval, PointerGetDatum(r),
					  entry->rel, entry->page,
					  entry->offset, sizeof(MACAKEY), FALSE);
	}
	else
		retval = entry;
	PG_RETURN_POINTER(retval);
}



Datum
gmaca_consistent(PG_FUNCTION_ARGS)
{
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
	macaddr    *query = (macaddr *) PG_GETARG_POINTER(1);
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);

	bool       retval;
	MACAKEY	   *key;

	if (!entry->key)
		return FALSE;
	key = (MACAKEY *) DatumGetPointer(entry->key);

	switch (strategy)
	{
		case BTLessEqualStrategyNumber:
			retval = MCGE(query, &(key->lower));
			break;
		case BTLessStrategyNumber:
			if (GIST_LEAF(entry))
				retval = MCGT(query, &(key->lower));
			else
				retval = MCGE(query, &(key->lower));
			break;
		case BTEqualStrategyNumber:
			/* in leaf page key->lower always = key->upper */
			if (GIST_LEAF(entry))
				retval = MCEQ(query, &(key->lower));
			else
				retval = (MCLE(&(key->lower), query) && MCLE(query, &(key->upper)));
			break;
		case BTGreaterStrategyNumber:
			if (GIST_LEAF(entry))
				retval = MCLT(query, &(key->upper));
			else
				retval = MCLE(query, &(key->upper));
			break;
		case BTGreaterEqualStrategyNumber:
			retval = MCLE(query, &(key->upper));
			break;
		default:
			retval = FALSE;
	}

	PG_RETURN_BOOL( retval );
}


static int64 mac_2_int64  ( macaddr * m ){
   unsigned char * mi = ( unsigned char * ) m;
   int64    res = 0;
   int i;
   for (i=0; i<6; i++ ){
     res += mi[i] << ( (5-i)*8 );  
   }
   return res;
}

static macaddr * int64_2_mac  ( macaddr * m , int64 * mi ){

   unsigned char * me = ( unsigned char * ) m;
   int              i ;
   for (i=0; i<6; i++ ){
     me[i] = ( *mi >> ( (5-i)*8 ) ) % 256 ;
   }
   return m;
}


Datum
gmaca_union(PG_FUNCTION_ARGS)
{
	bytea	   *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int			i,
				numranges;
	MACAKEY	   *cur,
		   *out = palloc(sizeof(MACAKEY));

        int64 iout[2], icur[2] ;

	numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	*(int *) PG_GETARG_POINTER(1) = sizeof(MACAKEY);

	cur = (MACAKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
        icur[0] =  mac_2_int64 ( &cur->lower );
        icur[1] =  mac_2_int64 ( &cur->upper );
        iout[0] = icur[0];
        iout[1] = icur[1];

	for (i = 1; i < numranges; i++)
	{
		cur = (MACAKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
                icur[0] =  mac_2_int64 ( &cur->lower );
                icur[1] =  mac_2_int64 ( &cur->upper );

		if ( iout[0] > icur[0] )
			iout[0] = icur[0];
		if ( iout[1] < icur[1] )
			iout[1] = icur[1];
	}
        int64_2_mac  ( &out->lower , &iout[0] );
        int64_2_mac  ( &out->upper , &iout[1] );
	PG_RETURN_POINTER(out);
}

Datum
gmaca_penalty(PG_FUNCTION_ARGS)
{
	MACAKEY	   *origentry = (MACAKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
	MACAKEY	   *newentry  = (MACAKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
	float	   *result    = (float *) PG_GETARG_POINTER(2);
        int64      iorg[2], inew[2] ;

        iorg[0] =  mac_2_int64 ( &origentry->lower );
        iorg[1] =  mac_2_int64 ( &origentry->upper );
        inew[0] =  mac_2_int64 ( &newentry->lower  );
        inew[1] =  mac_2_int64 ( &newentry->upper  );

        *result = Max(inew[1] - iorg[1], 0) +
                Max(iorg[0] - inew[0] , 0);

	PG_RETURN_POINTER(result);
}

Datum
gmaca_picksplit(PG_FUNCTION_ARGS)
{
	PG_RETURN_POINTER(btree_picksplit(
									  (bytea *) PG_GETARG_POINTER(0),
								  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
									  gmaca_binary_union,
									  mckey_cmp
									  ));
}

Datum
gmaca_same(PG_FUNCTION_ARGS)
{
	MACAKEY	   *b1 = (MACAKEY *) PG_GETARG_POINTER(0);
	MACAKEY	   *b2 = (MACAKEY *) PG_GETARG_POINTER(1);

	bool	   *result = (bool *) PG_GETARG_POINTER(2);

	if (b1 && b2)
		*result = (MCEQ(&(b1->lower), &(b2->lower)) && MCEQ(&(b1->upper), &(b2->upper))) ? TRUE : FALSE;
	else
		*result = (b1 == NULL && b2 == NULL) ? TRUE : FALSE;
	PG_RETURN_POINTER(result);
}

static void
gmaca_binary_union(Datum *r1, char *r2)
{
	MACAKEY	   *b1;
	MACAKEY	   *b2 = (MACAKEY *) r2;

	if (!DatumGetPointer(*r1))
	{
		*r1 = PointerGetDatum(palloc(sizeof(MACAKEY)));
		b1 = (MACAKEY *) DatumGetPointer(*r1);
		memcpy((void*) &b1->upper,(void*) &b2->upper,sizeof(macaddr) );
		memcpy((void*) &b1->lower,(void*) &b2->lower,sizeof(macaddr) );
	}
	else
	{
		b1 = (MACAKEY *) DatumGetPointer(*r1);
                if ( MCGT(&b1->lower, &b2->lower) ){
    		  memcpy((void*) &b1->lower,(void*) &b2->lower,sizeof(macaddr) );
                }
                if ( MCLE(&b1->upper, &b2->upper) ){
    		  memcpy((void*) &b1->upper,(void*) &b2->upper ,sizeof(macaddr) );
                }
	}
}

static int
mckey_cmp(const void *a, const void *b)
{
	return DatumGetInt32(
		 DirectFunctionCall2(
			 macaddr_cmp ,
			  PointerGetDatum (&(((MACAKEY *) (((RIX *) a)->r))->lower)),
			   PointerGetDatum (&(((MACAKEY *) (((RIX *) b)->r))->lower))
											 )
		);
}

